/*
 * This confidential and proprietary software may be used only as
 * authorised by a licensing agreement from ARM Limited
 * (C) COPYRIGHT 2012 ARM Limited
 * ALL RIGHTS RESERVED
 * The entire notice above must be reproduced on all authorised
 * copies and copies may only be made to the extent permitted
 * by a licensing agreement from ARM Limited.
 */

#ifndef TEXTURE_H
#define TEXTURE_H

#include "ETCHeader.h"

#if GLES_VERSION == 2
#include <GLES2/gl2.h>
#elif GLES_VERSION == 3
#include <GLES3/gl3.h>
#else 
#error "GLES_VERSION must be defined as either 2 or 3"
#endif

namespace MaliSDK
{
    /** 
     * \brief Functions for working with textures.
     */
    class Texture
    {
    private:
        /**
         * \brief Uses glGetIntegerv to get the number of compressed texture formats and the formats themselves.
         
         * Calls glGetIntegerv with GL_NUM_COMPRESSED_TEXTURE_FORMATS and GL_COMPRESSED_TEXTURE_FORMATS.
         * \param[out] textureFormats Pointer to the array of texture formats.
         * \param[out] numberOfTextureFormats Pointer to the number of compressed texture formats.
         */
        static void getCompressedTextureFormats(GLint **textureFormats, int* numberOfTextureFormats);
    public:
        /**
         * \brief Reports whether or not ETC (Ericsson Texture Compression) is supported.
         *
         * Uses getCompressedTextureFormats to get the list of supported compression 
         * formats and then checks to see if any of them are GL_ETC1_RGB8_OES.
         * \param[in] verbose If true, prints out the number of supported texture compression formats and then lists the formats supported.
         */
        static bool isETCSupported(bool verbose = false);

        /**
         * \brief Create a texture using random data.
         *     
         * \param[in] width The required width of the texture.
         * \param[in] height The required height of the texture.
         * \param[in] textureData A pointer to the created texture data.
         */
        static void createTexture(unsigned int width, unsigned int height, GLvoid **textureData);

        /**
         * \brief Create a 1 component texture of uniform colour.
         * 
         * \param [in]  width       The required width of the texture.
         * \param [in]  height      The required height of the texture.
         * \param [in]  red         The required red channel.
         * \param [out] textureData Output texture.
         */
        static void createTexture(unsigned int width, unsigned int height, unsigned int red, GLvoid **textureData);

        /**
         * \brief Deletes previously created texture.
         * 
         * \param [in] Texture to be deleted.
         */
        static void deleteTextureData(GLvoid **textureData);

        /**
         * \brief Load texture data from a file into memory.
         *
         * \param[in] filename The filename of the texture to load.
         * \param[out] textureData Pointer to the texture that has been loaded.
         */
        static void loadData(const char *filename, unsigned char **textureData);

        /**
         * \brief Load header and texture data from a pkm file into memory.
         *
         * \param[in] filename The filename of the texture to load.
         * \param[out] etcHeader Pointer to the header that has been loaded.
         * \param[out] textureData Pointer to the texture that has been loaded.      
         */
        static void loadPKMData(const char *filename, ETCHeader* etcHeader, unsigned char **textureData);

        /**
         * \brief Load compressed mipmaps into memory
         *
         * Load the base level, calculate how many Mipmap levels there are.
         * Load the PKM files into memory.
         * Load the data into the texture mipmap levels.
         * \param[in] filenameBase The base filename of the texture mipmap levels. Will have the mipmap level number appended to it to load all of the mipmap levels.
         * For example, if filenameBase = "texture_", this method will try to load the files "texture_0", "texture_1", "texture_2", etc..
         * \param[in] filenameSuffix Any suffix to the mipmap filenames. Most commonly used for file extensions.
         * For example, if filenameSuffix = ".pkm", this method will append ".pkm" to all the files it tries to load.
         * \param[out] textureID The texture ID of the texture that has been loaded.
         */
        static void loadCompressedMipmaps(const char *filenameBase, const char *filenameSuffix, GLuint *textureID);
    };
}
#endif /* TEXTURE_H */